#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, Sequence

from airflow.io.path import ObjectStoragePath
from airflow.models import BaseOperator

if TYPE_CHECKING:
    from airflow.providers.openlineage.extractors import OperatorLineage
    from airflow.utils.context import Context


class FileTransferOperator(BaseOperator):
    """
    Copies a file from a source to a destination.

    This streams the file from the source to the destination if required
    , so it does not need to fit into memory.

    :param src: The source file path or ObjectStoragePath object.
    :param dst: The destination file path or ObjectStoragePath object.
    :param source_conn_id: The optional source connection id.
    :param dest_conn_id: The optional destination connection id.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:FileTransferOperator`
    """

    template_fields: Sequence[str] = ("src", "dst")

    def __init__(
        self,
        *,
        src: str | ObjectStoragePath,
        dst: str | ObjectStoragePath,
        source_conn_id: str | None = None,
        dest_conn_id: str | None = None,
        overwrite: bool = False,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        self.src = src
        self.dst = dst
        self.source_conn_id = source_conn_id
        self.dst_conn_id = dest_conn_id
        self.overwrite = overwrite

    def execute(self, context: Context) -> None:
        src: ObjectStoragePath = self._get_path(self.src, self.source_conn_id)
        dst: ObjectStoragePath = self._get_path(self.dst, self.dst_conn_id)

        if not self.overwrite:
            if dst.exists() and dst.is_file():
                raise ValueError(f"Destination {dst} already exists")

        src.copy(dst)

    def get_openlineage_facets_on_start(self) -> OperatorLineage:
        from airflow.providers.common.compat.openlineage.facet import Dataset
        from airflow.providers.openlineage.extractors import OperatorLineage

        def _prepare_ol_dataset(path: ObjectStoragePath) -> Dataset:
            if hasattr(path, "namespace"):
                # namespace has been added in Airflow 2.9.0; #36410
                return Dataset(namespace=path.namespace, name=path.key)
            # manually recreating namespace
            return Dataset(
                namespace=f"{path.protocol}://{path.bucket}" if path.bucket else path.protocol,
                name=path.key.lstrip(path.sep),
            )

        src: ObjectStoragePath = self._get_path(self.src, self.source_conn_id)
        dst: ObjectStoragePath = self._get_path(self.dst, self.dst_conn_id)

        input_dataset = _prepare_ol_dataset(src)
        output_dataset = _prepare_ol_dataset(dst)

        return OperatorLineage(
            inputs=[input_dataset],
            outputs=[output_dataset],
        )

    @staticmethod
    def _get_path(path: str | ObjectStoragePath, conn_id: str | None) -> ObjectStoragePath:
        if isinstance(path, str):
            return ObjectStoragePath(path, conn_id=conn_id)
        return path



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
import json
import uuid
from functools import cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, TypeVar
from urllib.parse import urlsplit

import fsspec.utils

from airflow.configuration import conf
from airflow.io.path import ObjectStoragePath
from airflow.models.xcom import BaseXCom
from airflow.utils.json import XComDecoder, XComEncoder

if TYPE_CHECKING:
    from sqlalchemy.orm import Session

    from airflow.models import XCom

T = TypeVar("T")

SECTION = "common.io"


def _get_compression_suffix(compression: str) -> str:
    """
    Return the compression suffix for the given compression.

    :raises ValueError: if the compression is not supported
    """
    for suffix, c in fsspec.utils.compressions.items():
        if c == compression:
            return suffix

    raise ValueError(f"Compression {compression} is not supported. Make sure it is installed.")


@cache
def _get_base_path() -> ObjectStoragePath:
    return ObjectStoragePath(conf.get_mandatory_value(SECTION, "xcom_objectstorage_path"))


@cache
def _get_compression() -> str | None:
    return conf.get(SECTION, "xcom_objectstorage_compression", fallback=None) or None


@cache
def _get_threshold() -> int:
    return conf.getint(SECTION, "xcom_objectstorage_threshold", fallback=-1)


class XComObjectStorageBackend(BaseXCom):
    """
    XCom backend that stores data in an object store or database depending on the size of the data.

    If the value is larger than the configured threshold, it will be stored in an object store.
    Otherwise, it will be stored in the database. If it is stored in an object store, the path
    to the object in the store will be returned and saved in the database (by BaseXCom). Otherwise, the value
    itself will be returned and thus saved in the database.
    """

    @staticmethod
    def _get_full_path(data: str) -> ObjectStoragePath:
        """
        Get the path from stored value.

        :raises ValueError: if the key is not relative to the configured path
        :raises TypeError: if the url is not a valid url or cannot be split
        """
        p = _get_base_path()

        # normalize the path
        try:
            url = urlsplit(data)
        except AttributeError:
            raise TypeError(f"Not a valid url: {data}") from None

        if url.scheme:
            if not Path.is_relative_to(ObjectStoragePath(data), p):
                raise ValueError(f"Invalid key: {data}")
            return p / data.replace(str(p), "", 1).lstrip("/")

        raise ValueError(f"Not a valid url: {data}")

    @staticmethod
    def serialize_value(
        value: T,
        *,
        key: str | None = None,
        task_id: str | None = None,
        dag_id: str | None = None,
        run_id: str | None = None,
        map_index: int | None = None,
    ) -> bytes | str:
        # we will always serialize ourselves and not by BaseXCom as the deserialize method
        # from BaseXCom accepts only XCom objects and not the value directly
        s_val = json.dumps(value, cls=XComEncoder).encode("utf-8")

        if compression := _get_compression():
            suffix = f".{_get_compression_suffix(compression)}"
        else:
            suffix = ""

        threshold = _get_threshold()
        if threshold < 0 or len(s_val) < threshold:  # Either no threshold or value is small enough.
            return s_val

        base_path = _get_base_path()
        while True:  # Safeguard against collisions.
            p = base_path.joinpath(
                dag_id or "NO_DAG_ID",
                run_id or "NO_RUN_ID",
                task_id or "NO_TASK_ID",
                f"{uuid.uuid4()}{suffix}",
            )
            if not p.exists():
                break
        p.parent.mkdir(parents=True, exist_ok=True)

        with p.open(mode="wb", compression=compression) as f:
            f.write(s_val)
        return BaseXCom.serialize_value(str(p))

    @staticmethod
    def deserialize_value(result: XCom) -> Any:
        """
        Deserializes the value from the database or object storage.

        Compression is inferred from the file extension.
        """
        data = BaseXCom.deserialize_value(result)
        try:
            path = XComObjectStorageBackend._get_full_path(data)
        except (TypeError, ValueError):  # Likely value stored directly in the database.
            return data
        try:
            with path.open(mode="rb", compression="infer") as f:
                return json.load(f, cls=XComDecoder)
        except (TypeError, ValueError):
            return data

    @staticmethod
    def purge(xcom: XCom, session: Session) -> None:
        if not isinstance(xcom.value, str):
            return
        with contextlib.suppress(TypeError, ValueError):
            XComObjectStorageBackend._get_full_path(xcom.value).unlink(missing_ok=True)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# NOTE! THIS FILE IS AUTOMATICALLY GENERATED AND WILL BE
# OVERWRITTEN WHEN PREPARING PACKAGES.
#
# IF YOU WANT TO MODIFY THIS FILE, YOU SHOULD MODIFY THE TEMPLATE
# `get_provider_info_TEMPLATE.py.jinja2` IN the `dev/breeze/src/airflow_breeze/templates` DIRECTORY


def get_provider_info():
    return {
        "package-name": "apache-airflow-providers-common-io",
        "name": "Common IO",
        "description": "``Common IO Provider``\n",
        "state": "ready",
        "source-date-epoch": 1728484966,
        "versions": [
            "1.4.2",
            "1.4.1",
            "1.4.0",
            "1.3.2",
            "1.3.1",
            "1.3.0",
            "1.2.0",
            "1.1.0",
            "1.0.1",
            "1.0.0",
        ],
        "dependencies": ["apache-airflow>=2.8.0"],
        "integrations": [
            {
                "integration-name": "Common IO",
                "external-doc-url": "https://filesystem-spec.readthedocs.io/en/latest/index.html",
                "how-to-guide": ["/docs/apache-airflow-providers-common-io/operators.rst"],
                "tags": ["software"],
            }
        ],
        "operators": [
            {
                "integration-name": "Common IO",
                "python-modules": ["airflow.providers.common.io.operators.file_transfer"],
            }
        ],
        "xcom": ["airflow.providers.common.io.xcom.backend"],
        "asset-uris": [
            {
                "schemes": ["file"],
                "handler": "airflow.providers.common.io.assets.file.sanitize_uri",
                "to_openlineage_converter": "airflow.providers.common.io.assets.file.convert_asset_to_openlineage",
                "factory": "airflow.providers.common.io.assets.file.create_asset",
            }
        ],
        "dataset-uris": [
            {
                "schemes": ["file"],
                "handler": "airflow.providers.common.io.assets.file.sanitize_uri",
                "to_openlineage_converter": "airflow.providers.common.io.assets.file.convert_asset_to_openlineage",
                "factory": "airflow.providers.common.io.assets.file.create_asset",
            }
        ],
        "config": {
            "common.io": {
                "description": "Common IO configuration section",
                "options": {
                    "xcom_objectstorage_path": {
                        "description": "Path to a location on object storage where XComs can be stored in url format.\n",
                        "version_added": "1.3.0",
                        "type": "string",
                        "example": "s3://conn_id@bucket/path",
                        "default": "",
                    },
                    "xcom_objectstorage_threshold": {
                        "description": "Threshold in bytes for storing XComs in object storage. -1 means always store in the\ndatabase. 0 means always store in object storage. Any positive number means\nit will be stored in object storage if the size of the value is greater than the threshold.\n",
                        "version_added": "1.3.0",
                        "type": "integer",
                        "example": "1000000",
                        "default": "-1",
                    },
                    "xcom_objectstorage_compression": {
                        "description": "Compression algorithm to use when storing XComs in object storage. Supported algorithms\nare a.o.: snappy, zip, gzip, bz2, and lzma. If not specified, no compression will be used.\nNote that the compression algorithm must be available in the Python installation (e.g.\npython-snappy for snappy). Zip, gz, bz2 are available by default.\n",
                        "version_added": "1.3.0",
                        "type": "string",
                        "example": "gz",
                        "default": "",
                    },
                },
            }
        },
    }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import urllib.parse
from typing import TYPE_CHECKING

try:
    from airflow.assets import Asset
except ModuleNotFoundError:
    from airflow.datasets import Dataset as Asset  # type: ignore[no-redef]

if TYPE_CHECKING:
    from urllib.parse import SplitResult

    from airflow.providers.common.compat.openlineage.facet import Dataset as OpenLineageDataset


def create_asset(*, path: str, extra=None) -> Asset:
    # We assume that we get absolute path starting with /
    return Asset(uri=f"file://{path}", extra=extra)


def sanitize_uri(uri: SplitResult) -> SplitResult:
    if not uri.path:
        raise ValueError("URI format file:// must contain a non-empty path.")
    return uri


def convert_asset_to_openlineage(asset: Asset, lineage_context) -> OpenLineageDataset:
    """
    Translate Asset with valid AIP-60 uri to OpenLineage with assistance from the context.

    Windows paths are not standardized and can produce unexpected behaviour.
    """
    from airflow.providers.common.compat.openlineage.facet import Dataset as OpenLineageDataset

    parsed = urllib.parse.urlsplit(asset.uri)
    return OpenLineageDataset(namespace=f"file://{parsed.netloc}", name=parsed.path)
